#ifndef __KERNEL_H
#define __KERNEL_H

#include "config.h"
#include "vector.h"
#include "bv.h"

/* Ray Tracing Kernel functions
   and releted structures */

// roundoff tolerence
#define RAY_EPS 1e-6

// forward decs
class Object; // "object.h"
class Light; // "light.h"
class Colour; // "colour.h"
// data types

class Ray{
public:
	Vector o,d; // origin, direction (not necessarily normalized)
	Float t; // time of ray

	void bv_setup(); // sets up the optimisation info below.
	// call whenever ray has been changed
	Float p_dot_b[NUM_PSN];
	Float p_dot_a_rep[NUM_PSN]; // 1/(p_dot_a)
};

/*    Ray_Info
This structure is used by the kernel to store infomation about rays
being tested. It's members should be set to the appropriate info:
 o ray: set this to the ray that you want to be traced
 o generation: the generation of this ray. Users should increment this
   when appropriate. Initialized to 0 by K_primary_trace().
 o weight: weight of ray. Init to 1.0 by K_primary_trace().
 o type: type of ray. User should set this to reflected/transmitted (signifing
   a secondary ray). Set to primary by K_primary_trace().
 o object_in: object that ray is passing through. 0 if not passing through
   any object. (set to 0 by K_primary_trace())
*/
struct Ray_Info{
	Ray ray; // ray being considered
	int generation; // 0 for primary rays
	Float weight; // weight of this ray
	enum Ray_Type {primary,reflected,transmitted};
	Ray_Type type;
	Object* object_in;
};

/*    Intersect_Info
Used by kernel to return info about an intersection.
*/
struct Intersect_Info{
	Float t; // location on ray of intersection
	Object* obj; // who was hit
};	


// functions

void K_main(void); // should be called once. The kernel starts at this point

bool K_trace(Ray_Info&, Colour&);
/* K_trace()
This function will trace the ray specified and return true if it hit an
object in the scene. If it does hit an object, the appropriate colour is also
returned.
*/

bool K_primary_trace(Ray_Info&, Colour&);
/* K_primary_trace()
This function performs much the same task as K_trace(), but is used to signify
that this is a primary ray. Initalizes generation to 0, weight to 1.0,
 and starts keeping ray statistics.
*/

bool K_shadow_test(Ray&,Float,Light*);
/* K_shadow_test
Tests to see if there are any objects along the ray, the light being
the specified distance from the ray origin. returns true if the is
an obstacle
*/

Light* K_get_lights(); // returns the list of lights (can be used by shader)

Float K_random(); // returns random number [0,1)

#endif


