#include "linear.h"
#include <math.h>

int max_segments; // number of frames

void L_init(int nframes)
{
	max_segments=nframes;
}

static L_fn make_linear(Function*f)
{
	L_fn fn;
	fn.T0=f->t0;
	fn.T1=f->t1;
	fn.c=new L_coeffs[max_segments];

	double x0,x1,y0,y1;
	double segment_width=(fn.T1-fn.T0)/double(max_segments);
	for(int i=0;i<max_segments;i++){
		// compute line segment i
		x0=fn.T0+ i*segment_width;
		x1=fn.T0+ (i+1)*segment_width;
		y0=f->eval(x0);
		y1=f->eval(x1);
		fn.c[i].k0=(y1-y0)/(x1-x0);
		fn.c[i].k1=(x1*y0-x0*y1)/(x1-x0);
	}
	return fn;
}

static void make_linear(L_Matrix& M, TMatrix& tM)
{
	M.f[0]=make_linear(tM.r1);
	M.f[1]=make_linear(tM.r2);
	M.f[2]=make_linear(tM.r3);
	M.f[3]=make_linear(tM.r4);
	M.f[4]=make_linear(tM.r5);
	M.f[5]=make_linear(tM.r6);
	M.f[6]=make_linear(tM.r7);
	M.f[7]=make_linear(tM.r8);
	M.f[8]=make_linear(tM.r9);
	M.f[9]=make_linear(tM.tx);
	M.f[10]=make_linear(tM.ty);
	M.f[11]=make_linear(tM.tz);
}

L_TTransformation* L_linearize(TTransformation* T)
{
	if(T->is_linear()) return (L_TTransformation*)T;
	L_TTransformation* nT=new L_TTransformation(T->t0,T->t1);
	// ok, linearize each function in Ttrans and Tinverse
	// and put them into nT
	
	make_linear(nT->M,T->Ttrans);

	// inverse is calculated from M(t) by eval_inverse

	return nT;
}

L_TTransformation::L_TTransformation(Float a,Float b):
	TTransformation(a,b)
{}

// this is ment to get M=[A T] and return
// fn=P.A(t)(x,y,z,t) + P.T(t)
/*  this is

     [ px ] T [ a0 a1 a2 0 ] [ x ]    T [ tx ]
fn = [ py ]   [ a3 a4 a5 0 ] [ y ] + P  [ ty ]
     [ pz ]   [ a6 a7 a8 0 ] [ z ]      [ tz ]
     [ pt ]   [  0  0  0 1 ] [ t ]      [  0 ]

   =   Ax + By + Cz + pt*t + px.tx + py.ty + pz.tz
*/
L_proj_fn L_TTransformation::project(int i,Vector P, Float Pt)
{
	L_proj_fn fn;
	fn.A.k0=P.x*M.f[0].c[i].k0 + P.y*M.f[3].c[i].k0 + P.z*M.f[6].c[i].k0;
	fn.A.k1=P.x*M.f[0].c[i].k1 + P.y*M.f[3].c[i].k1 + P.z*M.f[6].c[i].k1;

	fn.B.k0=P.x*M.f[1].c[i].k0 + P.y*M.f[4].c[i].k0 + P.z*M.f[7].c[i].k0;
	fn.B.k1=P.x*M.f[1].c[i].k1 + P.y*M.f[4].c[i].k1 + P.z*M.f[7].c[i].k1;

	fn.C.k0=P.x*M.f[2].c[i].k0 + P.y*M.f[5].c[i].k0 + P.z*M.f[8].c[i].k0;
	fn.C.k1=P.x*M.f[2].c[i].k1 + P.y*M.f[5].c[i].k1 + P.z*M.f[8].c[i].k1;

	fn.D.k0=Pt+P.x*M.f[9].c[i].k0+P.y*M.f[10].c[i].k0+P.z*M.f[11].c[i].k0;
	fn.D.k1=   P.x*M.f[9].c[i].k1+P.y*M.f[10].c[i].k1+P.z*M.f[11].c[i].k1;

	return fn;
}

/*
	the inverse of
    [  A   T ]
M = [0 0 0 1 ]

is

M'= [ A'   -(A'T) ]
    [0 0 0    1   ]

 from matlab, the inverse of
                                 [ r0  r1  r2 ]
                                 [            ]
                                 [ r3  r4  r5 ]
                                 [            ]
                                 [ r6  r7  r8 ]

is
 
             [  r4 r8 - r5 r7     r1 r8 - r2 r7   r1 r5 - r2 r4  ]
             [  -------------   - -------------   -------------  ]
             [        %1                %1              %1       ]
             [                                                   ]
             [   r3 r8 - r5 r6   r0 r8 - r2 r6     r0 r5 - r2 r3 ]
             [ - -------------   -------------   - ------------- ]
             [         %1              %1                %1      ]
             [                                                   ]
             [  r3 r7 - r4 r6     r0 r7 - r1 r6   r0 r4 - r1 r3  ]
             [  -------------   - -------------   -------------  ]
             [        %1                %1              %1       ]
 
%1 :=  r0 r4 r8 - r0 r5 r7 - r3 r1 r8 + r3 r2 r7 + r6 r1 r5 - r6 r2 r4
*/
void L_TTransformation::eval_inverse(Float T)
{
	if(T==last_t) return;
	last_t=T;
	Float r[9];
	Float t[3];
	int i;
	int index=floor(max_segments*(T-t0)/(t1-t0));
	// eval M at t
	for(i=0;i<9;i++)
		r[i]=eval(M.f[i].c[index],T);
	for(i=0;i<3;i++) t[i]=eval(M.f[9+i].c[index],T);

	// compute 1.0/%1
	Float fact=1.0/(r[0]*r[4]*r[8] - r[0]*r[5]*r[7] - r[3]*r[1]*r[8] +
		   r[3]*r[2]*r[7] + r[6]*r[1]*r[5] - r[6]*r[2]*r[4]);

	// fill in inverse
	inverse.r1= (r[4]*r[8] - r[5]*r[7]) * fact;
	inverse.r2=-(r[1]*r[8] - r[2]*r[7]) * fact;
	inverse.r3= (r[1]*r[5] - r[2]*r[4]) * fact;

	inverse.r4=-(r[3]*r[8] - r[5]*r[6]) * fact;
	inverse.r5= (r[0]*r[8] - r[2]*r[6]) * fact;
	inverse.r6=-(r[0]*r[5] - r[2]*r[3]) * fact;

	inverse.r7= (r[3]*r[7] - r[4]*r[6]) * fact;
	inverse.r8=-(r[0]*r[7] - r[1]*r[6]) * fact;
	inverse.r9= (r[0]*r[4] - r[1]*r[3]) * fact;

	inverse.tx= -(inverse.r1*t[0] + inverse.r2*t[1] + inverse.r3*t[2]);
	inverse.ty= -(inverse.r4*t[0] + inverse.r5*t[1] + inverse.r6*t[2]);
	inverse.tz= -(inverse.r7*t[0] + inverse.r8*t[1] + inverse.r9*t[2]);
}

Vector L_TTransformation::apply(Float T)
{
	int index=floor(max_segments*(T-t0)/(t1-t0));
	Vector C;
	C.x=eval(M.f[9].c[index],T);
	C.y=eval(M.f[10].c[index],T);
	C.z=eval(M.f[11].c[index],T);
	return C;
}


