
#include "kernel.h"
#include "objectmanager.h"
#include "object.h"
#include "camera.h"
#include "light.h"
#include "stdlib.h"
/* implementation of Glassners hybrid SS/HVB technique */

/* notes:
o we assume world bounding box is square (and so all cells are too)
  because glassner84 said "parent node...encloses world sphere"

*/

// testing info
static Counter rayobject_tests=0; // number of ray/object intersections
static Counter rayobject_hits=0; // number of tests that hit
static Counter raybv_cell_tests=0;
static Counter raybv_subcell_tests=0;
static Counter ray_mailbox_hits=0;
static Counter num_leafs=0;
static Counter num_empty_leafs=0;
static Counter num_objs_in_leafs=0;
static Counter num_nodes=0;
static Counter max_octree_depth=0;
static Counter max_depth_reached=0;


void swap(int&a,int&b){
	int temp=a;
	a=b;
	b=temp;
}
void swap(Float&a,Float&b){
	Float temp=a;
	a=b;
	b=temp;
}
struct Vector4{ // 4D vector
	Float x,y,z,t;
};

struct HBV_node{
	BV bound;
	Vector4 pos; // corner of cell (in (-1,-1,-1,-1) direction)
	Float size; // size (ie width) of cell
	bool is_leaf; // true if this is a leaf, otherwise its a nonleaf
	union{
		HBV_node* children[16]; // we could make this dynamic to
				// save memory in leaves
		struct{
			Object** objs;
			BV* bvs;
			int nobjs;
		} leaf;
	};
};

#define TAB(x) {for(int i=0;i<x;i++) printf(" ");}

static HBV_node* Root_HBV;

static Object* obj_list;
static Light* light_list;

static Object** tmp_obj_arr; // this is inited in init() and used by create_hbv()
static BV* tmp_bv_arr;

/******* FUNCTIONS **********/

#define MAX_SUB_DEPTH 7
// returns true if se should subdivide the cell (pos+size) given the
// objects
bool need_to_subdivide(Vector4 pos,Float size,int depth,
	Object** objs,int nobjs)
{
	// stop if too deep
	if(depth>=MAX_SUB_DEPTH){
		max_depth_reached++;
		return false;
	}
	// subdivide if more then three objects in cell
	if(nobjs>3) return true;
	return false;
}

// create_HBV()
// given a list of objects, and a cell (defined by pos and size),
// this function returns a HBV within the cell, containing the
// appropriate objects from the list. depth could be used
// in the subdivision criteria
static HBV_node* create_HBV(Vector4 pos, Float size, int depth,
		Object** objects, int nobjs)
{
	// classify and create BV for each object within Box for this cell
	if(depth>max_octree_depth) max_octree_depth=depth;
	ST_Box box;
	box.x0=pos.x; box.y0=pos.y; box.z0=pos.z; box.t0=pos.t;
	box.x1=box.x0+size;
	box.y1=box.y0+size;
	box.z1=box.z0+size;
	box.t1=box.t0+size;
	int new_nobjs=0;
	BV bound;
	AA_Box aabox;
	for(int i=0;i<nobjs;i++){
		aabox.O=Vector(box.x0,box.y0,box.z0);
		aabox.X=Vector(box.x1,box.y0,box.z0);
		aabox.Y=Vector(box.x0,box.y1,box.z0);
		aabox.Z=Vector(box.x0,box.y0,box.z1);
		aabox.t0=box.t0;
		aabox.t1=box.t1;
		if(objects[i]->inside(aabox)){ // definitely a winner
			bound=objects[i]->bound;
			clip(bound,box);
			tmp_obj_arr[new_nobjs]=objects[i];
			tmp_bv_arr[new_nobjs]=bound;
			new_nobjs++;
		}
	}
	if(new_nobjs==0){
		num_empty_leafs++;
		return 0; // empty cell
	}
	// make local copy of remainin objects
	Object** new_objs=new Object*[new_nobjs];
	BV* new_bvs=new BV[new_nobjs];
if( !new_bvs || !new_objs){
printf("Out of memory: glassner.cc#1, depth=%d\n",depth);
exit(1);
}
	for(i=0;i<new_nobjs;i++){
		new_objs[i]=tmp_obj_arr[i];
		new_bvs[i]=tmp_bv_arr[i];
	}

	// create a new HBV
	HBV_node* HBV=new HBV_node;
	//HBV->BV= union of all child bvs
	HBV->bound=new_bvs[0];
	for(i=1;i<new_nobjs;i++){
		HBV->bound=BV_union(HBV->bound,new_bvs[i]);
	}
	HBV->pos=pos;
	HBV->size=size;

	if(need_to_subdivide(pos,size,depth,new_objs,new_nobjs)){
		num_nodes++;
		HBV->is_leaf=false;
		Float new_size=size/2.0;
		// need to subdivide further
		// for each sub-cell
		int child=0;
		for(int T=0;T<2;T++) for(int Z=0;Z<2;Z++) for(int Y=0;Y<2;Y++) for(int X=0;X<2;X++){
			Vector4 sub_pos=pos;
			sub_pos.x+=X*new_size;
			sub_pos.y+=Y*new_size;
			sub_pos.z+=Z*new_size;
			sub_pos.t+=T*new_size;
			HBV->children[child]=
			  create_HBV(sub_pos,new_size,depth+1,new_objs,new_nobjs);
			child++;
		}
		// if we only have one child, replace this HBV with that child
		child=0; // count
		HBV_node* child_node;
		for(int i=0;i<16;i++)
			if(HBV->children[i]!=0){
				child++;
				child_node=HBV->children[i];
			}
		if(child==1){
			delete HBV;
			num_nodes--;
			HBV=child_node;
		}
		delete[] new_objs;
		delete[] new_bvs;
	}
	else{ // a leaf node
		num_leafs++;
		num_objs_in_leafs+=new_nobjs;
		HBV->is_leaf=true;
		HBV->leaf.objs=new_objs;
		HBV->leaf.bvs=new_bvs;
		HBV->leaf.nobjs=new_nobjs;
	}
	return HBV;
}

void OM_init(Object* Optr,Light* Lptr)
{
	light_list=Lptr; obj_list=Optr;
	int nobjs=0; // count the number of objects
	for(Object* o_ptr=obj_list; o_ptr; o_ptr=o_ptr->next) nobjs++;
	// init temp array
	tmp_obj_arr=new Object*[nobjs];
	tmp_bv_arr=new BV[nobjs];
	// make our own object array
	Object** objs=new Object*[nobjs];
	int i;
	for(o_ptr=obj_list,i=0; o_ptr; o_ptr=o_ptr->next,i++)
		objs[i]=o_ptr;
	// find scene enclosing bound
	BV bv=objs[0]->bound;
	for(i=1;i<nobjs;i++)
		bv=BV_union(bv,objs[i]->bound);
	// enclose eyepoint
	//## do we need eyepoint?
	//bv=BV_union(bv,C_eyepoint());
	// find max side length and center of scene bv
	Vector4 P; Float size=0;
	Float widths[4];
	for(i=0;i<4;i++){
		widths[i]=bv.dfar[i]-bv.dnear[i];
		if(widths[i]>size) size=widths[i];
	}
	P.x=bv.dnear[0];//+ (widths[0]-size)*0.5;// ## which is better?
	P.y=bv.dnear[1];//+ (widths[1]-size)*0.5;
	P.z=bv.dnear[2];//+ (widths[2]-size)*0.5;
	P.t=bv.dnear[3];
	Root_HBV=create_HBV(P,size,0,objs,nobjs);
}

// these global variables are for recursive fn intersect_HBV()
static Ray ray;
static Object::MailBox ray_mailbox_no=1;
static Intersect_Info II;

// returns true if point is in or on cell of hbv
inline bool is_in(HBV_node* HBV, Vector p, Float time)
{
	// the ray_eps factor is need, especially for polygons on the edge
	// of the scene box.
	return  HBV->pos.x-RAY_EPS<=p.x && p.x<=HBV->pos.x+HBV->size+RAY_EPS &&
		HBV->pos.y-RAY_EPS<=p.y && p.y<=HBV->pos.y+HBV->size+RAY_EPS &&
		HBV->pos.z-RAY_EPS<=p.z && p.z<=HBV->pos.z+HBV->size+RAY_EPS &&
		HBV->pos.t<=time&&time<=HBV->pos.t+HBV->size ;
}
// this is a function used by intersect_HBV(). inlinned for speed
static inline void find_cell_path(HBV_node* HBV,int& num_crossings,int& current_cell,int* toggle_vals,Float* t_vals)
{
	Float Tm,Tp; // T- and T+
	Float Xm,Xp,Ym,Yp,Zm,Zp;
	Float Xd,Yd,Zd; // delta values

	float a,b;
	// find all the #- and #+
	if(ray.d.x==0.0){ Xm=-MAX_M; Xp=MAX_M;}
	else{
		a=(HBV->pos.x - ray.o.x)/ray.d.x;
		b=(HBV->pos.x+HBV->size - ray.o.x)/ray.d.x;
		if(a<b) {Xm=a; Xp=b;}
		else {Xm=b; Xp=a;}
	}
	if(ray.d.y==0.0){ Ym=-MAX_M; Yp=MAX_M;}
	else{
		a=(HBV->pos.y - ray.o.y)/ray.d.y;
		b=(HBV->pos.y+HBV->size - ray.o.y)/ray.d.y;
		if(a<b) {Ym=a; Yp=b;}
		else {Ym=b; Yp=a;}
	}
	if(ray.d.z==0.0){ Zm=-MAX_M; Zp=MAX_M;}
	else{
		a=(HBV->pos.z - ray.o.z)/ray.d.z;
		b=(HBV->pos.z+HBV->size - ray.o.z)/ray.d.z;
		if(a<b) {Zm=a; Zp=b;}
		else {Zm=b; Zp=a;}
	}
	// delta values
	Xd=(Xm+Xp)*0.5;
	Yd=(Ym+Yp)*0.5;
	Zd=(Zm+Zp)*0.5;
	// work out T- and T+
	Tm=Xm; Tp=Xp;
	if(Ym>Tm) Tm=Ym;
	if(Zm>Tm) Tm=Zm;
	if(Yp<Tp) Tp=Yp;
	if(Zp<Tp) Tp=Zp;
	if(Tm<0) Tm=0;
	// work out number of crossings, and fill in array
	num_crossings=0;
	t_vals[0]=-MAX_M;
	toggle_vals[1]=toggle_vals[2]=toggle_vals[3]=0;
	t_vals[1]=t_vals[2]=t_vals[3]=MAX_M;
	if(Tm<Xd && Xd<Tp){
		t_vals[1]=Xd;
		toggle_vals[1]=0x1;
		num_crossings=1;
	};
	if(Tm<Yd && Yd<Tp){
		t_vals[1+num_crossings]=Yd;
		toggle_vals[1+num_crossings]=0x2;
		num_crossings++;
	};
	if(Tm<Zd && Zd<Tp){
		t_vals[1+num_crossings]=Zd;
		toggle_vals[1+num_crossings]=0x4;
		num_crossings++;
	};
	// sort array ( an inline bubble sort)
	if(num_crossings>1){
		if(t_vals[2]<t_vals[1]) {// swap them
			swap(t_vals[1],t_vals[2]);
			swap(toggle_vals[1],toggle_vals[2]);
		}
		if(t_vals[3]<t_vals[2]) {// swap them
			swap(t_vals[2],t_vals[3]);
			swap(toggle_vals[2],toggle_vals[3]);
		}
		if(t_vals[2]<t_vals[1]) {// swap them
			swap(t_vals[1],t_vals[2]);
			swap(toggle_vals[1],toggle_vals[2]);
		}
	}
	// merge adjacent, equal cells
	if(t_vals[1]==t_vals[2] && num_crossings>1){
		toggle_vals[1]|=toggle_vals[2]; // merge
		t_vals[2]=t_vals[3]; // move next one up
		toggle_vals[2]=toggle_vals[3];
		num_crossings--;
	}
	if(t_vals[1]==t_vals[2] && num_crossings>1){
		toggle_vals[1]|=toggle_vals[2]; // merge
		t_vals[2]=t_vals[3]; // move next one up
		toggle_vals[2]=toggle_vals[3];
		num_crossings--;
	}
	// so now we have the middle of the arrays filled in,
	// with num_crossings entries.
	// fill in edge of arrays and work out which cell we start
	// in.
	t_vals[0]=Tm; t_vals[num_crossings+1]=Tp;
	Vector p=ray.o+ray.d*Tm;
	current_cell=0;
	Float g=HBV->pos.x+HBV->size*0.5;
	if(p.x>g || (p.x==g && ray.d.x>0) ) current_cell=0x1;
	g=HBV->pos.y+HBV->size*0.5;
	if(p.y>g || (p.y==g && ray.d.y>0) ) current_cell|=0x2;
	g=HBV->pos.z+HBV->size*0.5;
	if(p.z>g || (p.z==g && ray.d.z>0) ) current_cell|=0x4;
	if(ray.t>=HBV->pos.t+HBV->size*0.5) current_cell|=0x8;
}

bool intersect_HBV(HBV_node* HBV)
{
	if(HBV==0)
		return false; // empty child
	raybv_cell_tests++;
	if(!HBV->bound.intersect(ray)) //##is this worthwhile?
		return false; // missed bv
	if(! HBV->is_leaf){
		// find first cell that ray enters, and also ray path
		int num_crosses,current_cell;
		int toggle_vals[5]; // entries 0 and 4 aren't used
		Float t_vals[5];
		find_cell_path(HBV,num_crosses,current_cell,toggle_vals,t_vals);
		int index=0; // current index into t_vals
		do{
			index++;
			if(intersect_HBV(HBV->children[current_cell]))
				return true;
			// else go onto next cell
			current_cell^=toggle_vals[index]; // next cell
		}while(index<=num_crosses);
		return false; // ran out of this cell
	}
	else { // leaf node
		Float t,best_t=MAX_M;
		bool found=false;
		bool hit_obj=false;
		for(int i=0;i<HBV->leaf.nobjs;i++){ // for each object
			hit_obj=false;
			if(HBV->leaf.objs[i]->mailbox==ray_mailbox_no){
				ray_mailbox_hits++;
				hit_obj=true;
				t=HBV->leaf.objs[i]->mailbox_t;
			}
			else{ 
				raybv_subcell_tests++;
					if(HBV->leaf.bvs[i].intersect(ray)){
					rayobject_tests++;
					if(HBV->leaf.objs[i]->intersect(ray,t)){
						rayobject_hits++;
						HBV->leaf.objs[i]->mailbox=ray_mailbox_no;
						HBV->leaf.objs[i]->mailbox_t=t;
						hit_obj=true;
					}
				}
			}
			if(hit_obj && t<best_t){
				Vector p=ray.o+ray.d*t;
				// check to see if intersection is inside
				// cell
				if(is_in(HBV,p,ray.t)){
					found=true;
					best_t=t;
					II.t=t;
					II.obj=HBV->leaf.objs[i];
				}
			}
		}
		return found;
	}
}

bool OM_intersect(Ray_Info& ri,Intersect_Info& ii)
{
	ray=ri.ray;
	ray_mailbox_no++;
	if(intersect_HBV(Root_HBV)){
		ii=II;
		return true;
	}
	return false;
}

bool OM_shadow_test(Ray& r,Float d,Light*L)
{
	// test the cache
	Float t;
	if(L->last_blocked){
		if(L->last_blocked->intersect(ray,t) && t<d) return true;
	}

	ray=r;
	ray_mailbox_no++;
	if(intersect_HBV(Root_HBV)){
		if(II.t<d) return true;
	}
	return false;
}

void OM_starting_frame(int )
{
}

#include <iostream.h>

void OM_test_results(void)
{
	cerr<<"** Object Manager: Hybrid ASS/HBV\n";
	cerr<<"Ray/BV cell tests    :"<<raybv_cell_tests<<endl;
	cerr<<"Ray/BV subcell tests :"<<raybv_subcell_tests<<endl;
	cerr<<"Ray/Object tests     :"<<rayobject_tests<<endl;
	cerr<<"Ray/Object hits      :"<<rayobject_hits<<endl;
	cerr<<"Ray Mailbox hits     :"<<ray_mailbox_hits<<endl;
	cerr<<"Num Leaves           :"<<num_leafs<<endl;
	cerr<<"Num Empty Leaves     :"<<num_empty_leafs<<endl;
	cerr<<"Num non-leaf Nodes   :"<<num_nodes<<endl;
	cerr<<"Avg. Objs/leaf       :"<<double(num_objs_in_leafs)/num_leafs<<endl;
	cerr<<"Max octree depth     :"<<max_octree_depth<<" (!"<<max_depth_reached<<")"<<endl;
}


