/* my SpaceTime Ray Classification object manager */
#include "objectmanager.h"
#include "object.h"
#include "light.h"
#include "camera.h"
#include <iostream.h>
#include <stdlib.h>
#include "strcom.h"

static Object* obj_list;
static Light* light_list;

STRCinfo STRC={false,0};

// testing paramaters
static Counter rayobject_tests=0; // number of ray/object intersections
static Counter rayobject_hits=0; // number of tests that hit
static Counter raybv_tests=0;
static Counter classifications=0; // number of times the cand. set of a node was classified
static Counter subdivisions=0; // the number of subdivisions that took place
static Counter leaf_nodes=0; // number of leaf nodes
static Counter nodes_requested=0;
static Counter nodes_allocated=0;
static Counter non_time_subdivisions=0;
static Counter cache_checks=0;
static Counter cache_hits=0;

enum DomAxis {Xp=0,Xm=1,Yp=2,Ym=3,Zp=4,Zm=5};

struct point6D {
	Float x,y,z,t,u,v;
};

struct Node{ // a node in the tree
	Object** candidate_set;
	int nobjs; //number of objs in candidate set
	point6D pos,width;
	int depth;

	// type of node
	bool leaf; // is this a leaf node?
	bool subdivided; // if not a leaf, then has it been subdivided?
	/* leaf  subdivided  meaning
	    T     x           a leaf node
	    F     T           a node that has valid children
	    F     F           a node that hasn't been processed, no children
	*/
	bool t_divided; // true if it has been divided along t axis
	Node** children; // array of 2^6=64 (or 32 if not time divided) subnodes

	Node* next; // used to store unused nodes
};

Node axis_nodes[6];
BV world_bound;

// finds the min extent of this object along 'axis'
inline static Float min_extent(int axis, Object *obj)
{
	// uses knowlege about the order of vectors in
	// a bounding volume
	int vector=axis>>1; // X->0, Y->1, Z->2
	if(axis&1 == 0) // its a positive direction
		return obj->bound.dnear[vector];
	// else it is a minus direction
	return obj->bound.dfar[vector];
}

static void sort_objects(int axis,Object** IN,int nobjs,Object** OUT)
{
	// sort objects by their minimum extent along 'axis'
	// (use a insertion sort)
	
	for(int i=0;i<nobjs;i++) OUT[i]=IN[i];

	for(int i=0;i<nobjs-1;i++){
		int min_place=i;
		Float min_so_far=min_extent(axis,OUT[min_place]);
		for(int j=i+1;j<nobjs;j++){
			Float min=min_extent(axis,OUT[j]);
			if(min<min_so_far){
				min_so_far=min;
				min_place=i;
			}
		}
		if(i!=min_place){
			Object* tmp=OUT[i];
			OUT[i]=OUT[min_place];
			OUT[min_place]=tmp;
		}
	}
}

static Node* free_node_list=0; // a list of unused Nodes
inline static Node* new_node()
{
	nodes_requested++;
	if(free_node_list==0){
		nodes_allocated++;
		return new Node;
	}
	Node* n=free_node_list;
	free_node_list=n->next;
	return n;
}

inline static void delete_node(Node* node)
{
	node->next=free_node_list;
	free_node_list=node;
}

Node* node_cache[100];
DomAxis node_axis_cache[100];
Object** tmp_array;
void OM_init(Object* o, Light *l)
{
printf("N%d\n",sizeof(Node));
	int i;
	for(i=0;i<100;i++)node_cache[i]=0;
	obj_list=o;
	light_list=l;

	int nobjs=0; // count the number of objects
	Object* o_ptr;
	for(o_ptr=obj_list; o_ptr; o_ptr=o_ptr->next) nobjs++;

	// make an array copy of the objects
	Object**object_array=new Object*[nobjs];
	tmp_array=object_array;
	for(o_ptr=obj_list,i=0; o_ptr; o_ptr=o_ptr->next,i++)
		object_array[i]=o_ptr;

	// find scene enclosing bound
	world_bound=object_array[0]->bound;
	for(i=1;i<nobjs;i++)
		world_bound=BV_union(world_bound,object_array[i]->bound);

	// sort all the objects for each dominant axis, by their
	// minimum extent along that axis
	// also init each node
	point6D pos,width;
	pos.x=world_bound.dnear[0];
	pos.y=world_bound.dnear[1];
	pos.z=world_bound.dnear[2];
	pos.t=world_bound.dnear[3];
	pos.u=-1; pos.v=-1;
	width.x=world_bound.dfar[0]-world_bound.dnear[0];
	width.y=world_bound.dfar[1]-world_bound.dnear[1];
	width.z=world_bound.dfar[2]-world_bound.dnear[2];
	width.t=world_bound.dfar[3]-world_bound.dnear[3];
	width.u=2; width.v=2;
	for(int axis=0;axis<6;axis++){
		axis_nodes[axis].candidate_set=new Object*[nobjs];
		axis_nodes[axis].nobjs=nobjs;
		axis_nodes[axis].leaf=false;
		axis_nodes[axis].subdivided=false;
		axis_nodes[axis].pos=pos;
		axis_nodes[axis].width=width;
		axis_nodes[axis].depth=0;
		sort_objects(axis,object_array,nobjs,axis_nodes[axis].candidate_set);
	}
}

inline static void axis_of(Ray& r,DomAxis& axis, Float &U, Float &V)
{
	Float ax=ABS(r.d.x);
	Float ay=ABS(r.d.y);
	Float az=ABS(r.d.z);
	if( ax>ay && ax>az){ // X is dominant
		if(r.d.x>0) axis=Xp; else axis=Xm;
		U=r.d.y/ax;
		V=r.d.z/ax;
	}
	else if( ay>az){ // Y is dominant
		if(r.d.y>0) axis=Yp; else axis=Ym;
		U=r.d.x/ay;
		V=r.d.z/ay;
	}
	else { // Z is dominant
		if(r.d.z>0) axis=Zp; else axis=Zm;
		U=r.d.x/az;
		V=r.d.y/az;
	}
}

static point6D ray;
static DomAxis Axis;

inline static void make_test_planes(Node* node,Vector *N, Float *d, int& nplanes)
{
	nplanes=5;
	// bottom top left right
	Float x0,y0,z0,u0,v0;
	Float x1,y1,z1,u1,v1;
	x0=node->pos.x;
	y0=node->pos.y;
	z0=node->pos.z;
	u0=node->pos.u;
	v0=node->pos.v;
	x1=x0+node->width.x;
	y1=y0+node->width.y;
	z1=z0+node->width.z;
	u1=u0+node->width.u;
	v1=v0+node->width.v;
	switch(Axis){
	case Xp:
		//-X0.v0 +Z0   means -(x-x0)*v0 + (z-z0) = 0 is eqn of plane
		N[0]=Vector(-v0,0,1); d[0]=x0*v0-z0;
		//+X0.v1 -Z1
		N[1]=Vector(v1,0,-1); d[1]=-x0*v1+z1;
		//-X0.u0 +Y0
		N[2]=Vector(-u0,1,0); d[2]=x0*u0-y0;
		//+X0.u1 -Y1
		N[3]=Vector(u1,-1,0); d[3]=-x0*u1+y1;
		// back plane
		N[4]=Vector(1,0,0); d[4]=-x0;
		// other planes
		//if(v0>0) {N[nplanes]=Vector(0,0,1); d[nplanes++]=-1;}
		//else if(v1<0) {N[nplanes]=Vector(0,0,-1); d[nplanes++]=1;}
		//if(u0>0) {N[nplanes]=Vector(0,1,0); d[nplanes++]=-1;}
		//else if(u1<0) {N[nplanes]=Vector(0,0,-1); d[nplanes++]=1;}
		break;
	case Xm:
		//+X1.v0 +Z0
		N[0]=Vector(v0,0,1); d[0]=-x1*v0-z0;
		//-X1.v1 -Z1
		N[1]=Vector(-v1,0,-1); d[1]=x1*v1+z1;
		//-X1.u1 -Y1
		N[2]=Vector(-u1,-1,0); d[2]=x1*u1+y1;
		//+X1.u0 +Y0
		N[3]=Vector(u0,1,0); d[3]=-x1*u0-y0;
		// back plane
		N[4]=Vector(-1,0,0); d[4]=x1;
		// other planes
		//if(v0>0) {N[nplanes]=Vector(0,0,1); d[nplanes++]=-1;}
		//else if(v1<0) {N[nplanes]=Vector(0,0,-1); d[nplanes++]=1;}
		//if(u0>0) {N[nplanes]=Vector(0,1,0); d[nplanes++]=-1;}
		//else if(u1<0) {N[nplanes]=Vector(0,0,-1); d[nplanes++]=1;}
		break;
	case Yp:
		//-Y0*v0 +Z0
		N[0]=Vector(0,-v0,1); d[0]=y0*v0-z0;
		//+Y0*v1 -Z1
		N[1]=Vector(0,v1,-1); d[1]=-y0*v1+z1;
		//+Y0*u0 -X1
		N[2]=Vector(-1,u0,0); d[2]=-y0*u0+x1;
		//-Y0*u1 +X0
		N[3]=Vector(1,-u1,0); d[3]=y0*u1-x0;
		// back plane
		N[4]=Vector(0,1,0); d[4]=-y0;
		// other planes
		//if(v0>0) {N[nplanes]=Vector(0,0,1); d[nplanes++]=-1;}
		//else if(v1<0) {N[nplanes]=Vector(0,0,-1); d[nplanes++]=1;}
		//if(u0>0) {N[nplanes]=Vector(1,0,0); d[nplanes++]=-1;}
		//else if(u1<0) {N[nplanes]=Vector(-1,0,0); d[nplanes++]=1;}
		break;
	case Ym:
		//+Y1*v0 +Z0
		N[0]=Vector(0,v0,1); d[0]=-y1*v0-z0;
		//-Y1*v1 -Z1
		N[1]=Vector(0,-v1,-1); d[1]=y1*v1+z1;
		//+Y1*u1 +X0
		N[2]=Vector(1,u1,0); d[2]=-y1*u1-x0;
		//-Y1*u0 -X1
		N[3]=Vector(-1,-u0,0); d[3]=y1*u0+x1;
		// back plane
		N[4]=Vector(0,-1,0); d[4]=y1;
		// other planes
		//if(v0>0) {N[nplanes]=Vector(0,0,1); d[nplanes++]=-1;}
		//else if(v1<0) {N[nplanes]=Vector(0,0,-1); d[nplanes++]=1;}
		//if(u0>0) {N[nplanes]=Vector(1,0,0); d[nplanes++]=-1;}
		//else if(u1<0) {N[nplanes]=Vector(-1,0,0); d[nplanes++]=1;}
		break;
	case Zp:
		//-Z0*v0+Y0
		N[0]=Vector(0,1,-v0); d[0]=z0*v0-y0;
		//+Z0*v1-Y1
		N[1]=Vector(0,-1,v1); d[1]=-z0*v1+y1;
		//-Z0*u0+X0
		N[2]=Vector(1,0,-u0); d[2]=z0*u0-x0;
		//+Z0*u1-X1
		N[3]=Vector(-1,0,u1); d[3]=-z0*u1+x1;
		// back plane
		N[4]=Vector(0,0,1); d[4]=-z0;
		// other planes
		//if(v0>0) {N[nplanes]=Vector(0,1,0); d[nplanes++]=-1;}
		//else if(v1<0) {N[nplanes]=Vector(0,-1,0); d[nplanes++]=1;}
		//if(u0>0) {N[nplanes]=Vector(1,0,0); d[nplanes++]=-1;}
		//else if(u1<0) {N[nplanes]=Vector(-1,0,0); d[nplanes++]=1;}
		break;
	case Zm:
		//+Z1*v0 +Y0
		N[0]=Vector(0,1,v0); d[0]=-z1*v0-y0;
		//-Z1*v1 -Y1
		N[1]=Vector(0,-1,-v1); d[1]=z1*v1+y1;
		//-Z1*u1 -X1
		N[2]=Vector(-1,0,-u1); d[2]=z1*u1+x1;
		//+Z1*u0 +X0
		N[3]=Vector(1,0,u1); d[3]=-z1*u0-x0;
		// back plane
		N[4]=Vector(0,0,-1); d[4]=z1;
		// other planes
		//if(v0>0) {N[nplanes]=Vector(0,1,0); d[nplanes++]=-1;}
		//else if(v1<0) {N[nplanes]=Vector(0,-1,0); d[nplanes++]=1;}
		//if(u0>0) {N[nplanes]=Vector(1,0,0); d[nplanes++]=-1;}
		//else if(u1<0) {N[nplanes]=Vector(-1,0,0); d[nplanes++]=1;}
		break;
	}
	
}

static inline bool classify(Node* node,Object* obj, Vector *N, Float *d, int nplanes)
{
	if(obj->bound.dnear[3]>node->pos.t+node->width.t || obj->bound.dfar[3]<node->pos.t)
		return false;
	// can reject this object if all points of bounding box
	// are in -ve half space of any of the planes
	Float x0,y0,z0,x1,y1,z1;
	x0=obj->bound.dnear[0];
	y0=obj->bound.dnear[1];
	z0=obj->bound.dnear[2];
	x1=obj->bound.dfar[0];
	y1=obj->bound.dfar[1];
	z1=obj->bound.dfar[2];
	for(int i=0;i<nplanes;i++){
		if(N[i]*Vector(x0,y0,z0)+d[i]<0 &&
		   N[i]*Vector(x1,y0,z0)+d[i]<0 &&
		   N[i]*Vector(x0,y1,z0)+d[i]<0 &&
		   N[i]*Vector(x1,y1,z0)+d[i]<0 &&
		   N[i]*Vector(x0,y0,z1)+d[i]<0 &&
		   N[i]*Vector(x1,y0,z1)+d[i]<0 &&
		   N[i]*Vector(x0,y1,z1)+d[i]<0 &&
		   N[i]*Vector(x1,y1,z1)+d[i]<0 )
			return false;
	}
	return true;
}

Node* classify_ray(Node* node)
{
	// this is done iteritevly, not recursivly
	if(node==0 ) return 0;
	while(!node->leaf){
		if(node->subdivided){ // traverse to one of its children
			// work out which one
			int sub_node=0;
			if(ray.x>node->pos.x+node->width.x*0.5) sub_node|=0x01;
			if(ray.y>node->pos.y+node->width.y*0.5) sub_node|=0x02;
			if(ray.z>node->pos.z+node->width.z*0.5) sub_node|=0x04;
			if(ray.u>node->pos.u+node->width.u*0.5) sub_node|=0x08;
			if(ray.v>node->pos.v+node->width.v*0.5) sub_node|=0x10;
			if(node->t_divided)
			  if(ray.t>node->pos.t+node->width.t*0.5) sub_node|=0x20;
			node=node->children[sub_node];
		}
		else{
			// must process this node.
			// re-classify candidate set
			Vector N[9]; Float d[9]; int nplanes;
			classifications++;
			make_test_planes(node,N,d,nplanes);
			int j=0;
			for(int i=0;i<node->nobjs;i++){
				if(classify(node,node->candidate_set[i],N,d,nplanes))
					tmp_array[j++]=node->candidate_set[i];
			}
			node->candidate_set=new Object*[j];
			node->nobjs=j;
			int moving_count=0;
			for(int i=0;i<j;i++){
				node->candidate_set[i]=tmp_array[i];
				if(tmp_array[i]->type==Object::ttransform) moving_count++;
			}

			// evaluate subdivision critera
			if(j>6 && node->depth<5){ //##
				// subdivide
				subdivisions++;
				int num_children=64;
				node->t_divided=true;
				node->subdivided=true;
				if(STRC.smart_time_divide && moving_count<=STRC.num_moving){
					num_children=32;
					node->t_divided=false;
					non_time_subdivisions++;
				}
				node->children=new Node*[num_children];
				int i;
				for(i=0;i<num_children;i++) node->children[i]=new_node();
				i=0;
				point6D width=node->width;
				width.x*=0.5;
				width.y*=0.5;
				width.z*=0.5;
				width.u*=0.5;
				width.v*=0.5;
				if(node->t_divided) width.t*=0.5;
				point6D pos;
				int X,Y,Z,T,U,V;
				int Tend= node->t_divided? 2 : 1;
				for(T=0,pos.t=node->pos.t ;T<Tend; T++,pos.t+=width.t)
				for(V=0,pos.v=node->pos.v ;V<2; V++,pos.v+=width.v)
				for(U=0,pos.u=node->pos.u ;U<2; U++,pos.u+=width.u)
				for(Z=0,pos.z=node->pos.z ;Z<2; Z++,pos.z+=width.z)
				for(Y=0,pos.y=node->pos.y ;Y<2; Y++,pos.y+=width.y)
				for(X=0,pos.x=node->pos.x ;X<2; X++,pos.x+=width.x){
					Node *sub_node=node->children[i++];
					sub_node->candidate_set=node->candidate_set;
					sub_node->nobjs=node->nobjs;;
					sub_node->pos=pos;
					sub_node->width=width;
					sub_node->depth=node->depth+1;
					sub_node->leaf=false;
					sub_node->subdivided=false;
					sub_node->t_divided=node->t_divided;
				}
			}
			else { // tag as leaf
				leaf_nodes++;
				node->leaf=true;
			}
		}
	}
	return node;
}

bool OM_intersect(Ray_Info& ri ,Intersect_Info& ii)
{
	// check for intersection with world bound, and
	// reclassify origin
	Float t;
	if(!world_bound.intersect(ri.ray,t)) return false;
	if(t<0.0){ // ray's origin is in world bound
		ray.x=ri.ray.o.x;
		ray.y=ri.ray.o.y;
		ray.z=ri.ray.o.z;
	}
	else{ // push ray up to world bound
		Vector p=ri.ray.o+ri.ray.d*t;
		ray.x=p.x;
		ray.y=p.y;
		ray.z=p.z;
	}
	ray.t=ri.ray.t;
	axis_of(ri.ray,Axis,ray.u,ray.v);

	Node* node;
	// do cache check here 
	if(node_cache[ri.generation]){
		cache_checks++;
		node=node_cache[ri.generation];
		if( node->pos.x<=ray.x && ray.x<=node->pos.x+node->width.x &&
		    node->pos.y<=ray.y && ray.y<=node->pos.y+node->width.y &&
		    node->pos.z<=ray.z && ray.z<=node->pos.z+node->width.z &&
		    node->pos.u<=ray.u && ray.u<=node->pos.u+node->width.u &&
		    node->pos.v<=ray.v && ray.v<=node->pos.v+node->width.v &&
		    ( !node->t_divided ||
		      (node->pos.t<=ray.t && ray.t<=node->pos.t+node->width.t ))
		) cache_hits++; // cache hit
		else node=classify_ray(&axis_nodes[Axis]); // cache miss
	}
	else node=classify_ray(&axis_nodes[Axis]); // no cache entry
	node_cache[ri.generation]=node;

	if(node==0){
		cerr<<"Found a pruned node"<<endl; // should not happen
		exit(1);
	}

	// process candidate set
	Object** objs=node->candidate_set;
	Object* Obj=0; // closest object intersected
	Float T; // point of closest intersection

	// find first object intersected in candidate set
	int i;
	for(i=0;i<node->nobjs;i++){
		raybv_tests++;
		if(objs[i]->bound.intersect(ri.ray)){
			rayobject_tests++;
			if(objs[i]->intersect(ri.ray,T)){
				rayobject_hits++;
				Obj=objs[i];
				break;
			}
		}
	}
	if(Obj==0) return false; // couldn't find an intersection
	// now find any better intersection,
	// but stop when extent of objects lie beyond known intersection
	for(i++;i<node->nobjs;i++){
		if(min_extent(Axis,objs[i])>T) break; //##
		raybv_tests++;
		if(objs[i]->bound.intersect(ri.ray)){
			rayobject_tests++;
			if(objs[i]->intersect(ri.ray,t)){
				rayobject_hits++;
				if(t<T){
					T=t;
					Obj=objs[i];
				}
			}
		}
	}
	ii.t=T;
	ii.obj=Obj;
	return true;
}


bool OM_shadow_test(Ray& r, Float d,Light* L)
{
	// assume that ray origin is inside world bound
	ray.x=r.o.x;
	ray.y=r.o.y;
	ray.z=r.o.z;
	ray.t=r.t;
	axis_of(r,Axis,ray.u,ray.v);
	// no cache for shadows
	Node* node=classify_ray(&axis_nodes[Axis]); 

	// process candidate set
	Object** objs=node->candidate_set;
	Float T;

	// find first object intersected in candidate set
	for(int i=0;i<node->nobjs;i++){
		raybv_tests++;
		if(objs[i]->bound.intersect(r)){
			rayobject_tests++;
			if(objs[i]->intersect(r,T)){
				rayobject_hits++;
				if(T<d) return true;
			}
		}
	}
	return false;
}

// prunes brances of tree that are completely before given time
static void prune_before_time(Node* node, Float time)
{
	if(node->pos.t >= time) return; // no pruning possible
	if(node->leaf || !node->subdivided); // do nothing
	else{ // prune subtrees
		int num_children= node->t_divided?64:32;
		for(int i=0;i<num_children;i++){
			Node* subnode= node->children[i];
			if(subnode==0 || !subnode->t_divided) continue;
			// recurse this child if neccessary
			if(!subnode->leaf && subnode->subdivided)
				prune_before_time(subnode,time);
			if(subnode->pos.t+subnode->width.t<time){ //delete this child
				delete_node(subnode);
				node->children[i]=0;
			}
		}
	}
}

void OM_starting_frame(int frame)
{
	Float time=C_query_frame_time(frame);
	for(int i=0;i<6;i++)
		prune_before_time(&axis_nodes[i],time);
}

void OM_test_results(void)
{
	cerr<<"** Object Manager: STRC\n";
	cerr<<"Ray/BV tests     :"<<raybv_tests<<endl;
	cerr<<"Ray/Object tests :"<<rayobject_tests<<endl;
	cerr<<"Ray/Object hits  :"<<rayobject_hits<<endl;
	cerr<<"Classifications  :"<<classifications<<endl;
	cerr<<"Subdivisions     :"<<subdivisions<<endl;
	cerr<<"non-time subdivs.:"<<non_time_subdivisions<<endl;
	cerr<<"Leaf Nodes       :"<<leaf_nodes<<endl;
	cerr<<"Nodes requested  :"<<nodes_requested<<endl;
	cerr<<"Nodes allocated  :"<<nodes_allocated<<endl;
	cerr<<"Cache hit rate   :"<<100.0*(double(cache_hits)/double(cache_checks));
	cerr<<"% ("<<cache_hits<<"/"<<cache_checks<<")"<<endl;
}

