#ifndef __TRANSFORM_H
#define __TRANSFORM_H

#include "kernel.h"
#include "vector.h"
#include "object.h"
#include "tvector.h"

class Source_Generator;

/*
  N O T E S   A B O U T   T R A N S F O R M A T I O N S

We will use the mathematical notation from "CG: principles and practice",
Foley&VanDam. That is, a point is specified my a vector
  v = (x,y,z)
  and an affine transformation by
    [ r1 r2 r3 tx ]
    [ r4 r5 r6 ty ]
M = [ r7 r8 r9 tz ] = [   R   T ]
    [ 0  0  0  1  ]   [ 0 0 0 1 ]

An affine transformation is any combination of rotations, scales, shears
and translations. We apply a transformation to a point with:

 v' = M . v 

The compostion of the transformation M1 followed by
M2 is
  M' = M2 M1
so we have
  v' = (M2 M1) . v

The inverse transformation M~ of
  M = M1 M2 M3 ... Mn
is
  M~ = Mn~ ... M3~ M2~ M1~

*/


/* Matrix class. This doesn't represent a full matrix, just an
affine transformation as in M above. */

class Matrix{
public:
	Matrix();
	Float r1,r2,r3,r4,r5,r6,r7,r8,r9,tx,ty,tz;
};

// matrix multiply
Matrix operator* (const Matrix& a, const Matrix& b);

/* Transformation class. This represents a transformation, and
   also keeps track of the inverse transformation. It also has
   member functions that apply (basic) transformations to the
   current transformation */
class Transformation{
public:
	Matrix trans,inverse;
	void rotateX(Float);
	void rotateY(Float);
	void rotateZ(Float);
	void translate(Vector);
	void scale(Vector);
	void post_mult(Transformation*T);// computes (this)*(T)
	//void shear; ?? what are the matixes for these 
};

// transform vectors ## make these inline??
inline Vector transform_point(const Matrix& M, const Vector& b)
{
	Vector a;
	a.x=M.r1*b.x + M.r2*b.y + M.r3*b.z + M.tx;
	a.y=M.r4*b.x + M.r5*b.y + M.r6*b.z + M.ty;
	a.z=M.r7*b.x + M.r8*b.y + M.r9*b.z + M.tz;
	return a;
}
inline Vector transform_normal(const Matrix& M, const Vector& b)
{
	Vector a;
	a.x=M.r1*b.x + M.r2*b.y + M.r3*b.z;
	a.y=M.r4*b.x + M.r5*b.y + M.r6*b.z;
	a.z=M.r7*b.x + M.r8*b.y + M.r9*b.z;
	return a;
}
Vector transform_normal(const Matrix& a, const Vector& b);
Vector transform_normalT(const Matrix& a, const Vector& b);//transpose

// actually create objects with a transformation
// NOTE: only a shallow copy of Transformation is made!

Object* Transform_Object(Object*,Transformation*,Surface_Info&);
inline Object* Transform_Object(Object* o,Transformation* t)
	{ return Transform_Object(o,t,o->surface);}

class Transformed_Object: public Object{
public:
	Object* obj;
	Transformation& T;
	Transformed_Object(Object* o,Transformation* t,Surface_Info &s);
	virtual bool intersect(Ray&,Float&);
	virtual void aux_info(Ray&,Intersect_Info&,Shade_Info&);
	virtual void compute_bounds(BV&);
	virtual bool inside(AA_Box&);
};

#endif

